IMPLEMENTATION MODULE ConfigFile;

(*
 *    01.10.95 AM    -  in Dispose cfgVars freigeben
 *
 *    19.01.1995        Version von ConfigFile mit Debugoption
 *    10.10.1996     -  Diverse Fehler in der Speicherverwaltung beseitigt
 *    16.10.1996     -  benutzt ModReg.Register
 *)


FROM     System            IMPORT   GetEnv;
FROM     Storage           IMPORT   ALLOCATE, DEALLOCATE;
FROM     FileIO            IMPORT   File, AccessMode, Open, Close, EOF, ReadLine;
FROM     FileSpec          IMPORT   ExtractDirPath, ExtractFileName;
FROM     Directories       IMPORT   GetCurDir;
FROM     Paths             IMPORT   Locate;
IMPORT   InOut;
FROM     RealConversions   IMPORT   StringToReal;
FROM     NumberConversion  IMPORT   StringToInt, StringToCard, StringToLongInt;
FROM     Strings           IMPORT   CompareStr, Copy, Delete, Insert, Pos,
                                    Length, Append;
FROM     Strings1          IMPORT   RemoveLeadingSpaces, RemoveTrailingSpaces,
                                    Assign, Trim, Upper, CorrectPath;
IMPORT   ModReg;
IMPORT   void;


CONST    LINELEN     = 256;
         VARLEN      =  80;
         ENVLEN      =  80;

TYPE     ErrorType   =  (errZuweisungFehlt, errVarnameZuLang, errEnvVarZuLang,
                         errLeereZuweisung, errOutOfMem, errVarExists,
                         errWrongType, errNoReal, errNoInteger, errNoCardinal,
                         errNoLongint, errNoBoolean, errNoChar, errNoString,
                         errNoPath);

         VarType     =  (vUndef, vReal, vInteger, vCardinal, vLongint, vBoolean,
                         vChar, vString, vPath);

         Variable    =  POINTER TO VarRec;

         VarRec      =  RECORD
                           namelen: CARDINAL;
                           name:    POINTER TO ARRAY [0 .. VARLEN-1] OF CHAR;
                           valuelen: CARDINAL;
                           value:   POINTER TO ARRAY [0 .. LINELEN-1] OF CHAR;
                           default: RECORD
                                    CASE type: VarType OF
                                    | vUndef:
                                       undefined:  BOOLEAN;
                                    | vReal:
                                       real:       REAL;
                                    | vInteger:
                                       integer:    INTEGER;
                                    | vCardinal:
                                       cardinal:   CARDINAL;
                                    | vLongint:
                                       longint:    LONGINT;
                                    | vBoolean:
                                       boolean:    BOOLEAN;
                                    | vChar:
                                       char:       CHAR;
                                    | vString, vPath:
                                       size:       CARDINAL;
                                       string:     POINTER TO
                                                   ARRAY [0 .. LINELEN-1] OF CHAR;
                                    END;
                                    END;
                           next:    Variable;
                        END;

VAR   requestedPath,
      cfgPath:       ARRAY [0 .. 127] OF CHAR;
      cfgName:       ARRAY [0 .. 31] OF CHAR;
      cfgOpen:       BOOLEAN;
      cfgVars:       Variable;
      debug:         BOOLEAN;


PROCEDURE RemoveQuotes(VAR line: ARRAY OF CHAR;
                           null: BOOLEAN);
VAR   i, k,
      len,
      inc:  CARDINAL;
BEGIN
   (*debug
   InOut.WriteString("mit Quotes:  "); InOut.WriteLine(line);
   debug*)
   IF null THEN
      len := ORD(line[0]);
      inc := 1;
   ELSE
      len := Length(line);
      inc := 0;
   END;
   i := inc;
   WHILE i<(len+inc) DO
      IF line[i]="'" THEN
         (* Escapen _nicht_ mglich
         IF line[i+1]="'" THEN
            INC(i);
         END;
         *)
         FOR k := i TO len-1+inc DO
            line[k] := line[k+1];
         END;
         DEC(len);
      END;
      INC(i);
   END;
   (*debug
   InOut.WriteString("ohne Quotes: "); InOut.WriteLine(line);
   debug*)
END RemoveQuotes;

PROCEDURE Find(varname:    ARRAY OF CHAR): Variable;
(*
 * wenn varname definiert ist liefert die Funktion einen Zeiger auf die
 * zugehrigen Daten, sonst NIL
 *)
VAR   result,
      next:    Variable;
      found:   BOOLEAN;
BEGIN
   result := cfgVars;
   found := FALSE;
   next := result^.next;
   WHILE (NOT found) AND (next#NIL) DO
      result := next;
      (*debug
      InOut.WriteString(result^.name^); InOut.WriteString(" =?= ");
      InOut.WriteLine(varname);
      debug*)
      found := (CompareStr(result^.name^, varname)=0);
      next := result^.next;
   END;
   IF NOT found THEN
      result := NIL;
   END;
   RETURN result;
END Find;

PROCEDURE Error(err: ErrorType;
                str: ARRAY OF CHAR;
                in:  ARRAY OF CHAR);
BEGIN
   InOut.WriteString("Fehler in ");
   InOut.WriteString(in);
   InOut.WriteString(": ");
   CASE err OF
   | errZuweisungFehlt:
      InOut.WriteString("Zuweisung fehlt: ");
   | errVarnameZuLang:
      InOut.WriteString("Variablenname zu lang: ");
   | errEnvVarZuLang:
      InOut.WriteString("Name einer Environmentvariable zu lang: ");
   | errLeereZuweisung:
      InOut.WriteString("Leere Zuweisung: ");
   | errOutOfMem:
      InOut.WriteString("Speicher voll!");
   | errVarExists:
      InOut.WriteString("Variable bereits deklariert: ");
   | errWrongType:
      InOut.WriteString("Falscher Variablentyp angefordert: ");
   | errNoReal:
      InOut.WriteString("Angeforderte Variable enthlt keine REAL-Zahl: ");
   | errNoInteger:
      InOut.WriteString("Angeforderte Variable enthlt keine INTEGER-Zahl: ");
   | errNoCardinal:
      InOut.WriteString("Angeforderte Variable enthlt kein WORD: ");
   | errNoLongint:
      InOut.WriteString("Angeforderte Variable enthlt keine LONGINT-Zahl: ");
   | errNoBoolean:
      InOut.WriteString("Angeforderte Variable enthlt keinen BOOLEAN-Wert: ");
   | errNoChar:
      InOut.WriteString("Angeforderte Variable enthlt keinen CHAR-Wert: ");
   | errNoString: (* nur der Vollstndigkeit halber - kann gar nicht sein :) *)
      InOut.WriteString("Angeforderte Variable enthlt keinen String: ");
   | errNoPath:
      InOut.WriteString("Angeforderte Variable enthlt keinen (legalen) Pfad: ");
   ELSE
   END;
   IF str[0]#0C THEN
      InOut.WriteString(str);
   END;
   InOut.WriteLn;
END Error;


PROCEDURE Read(fileName:   ARRAY OF CHAR): BOOLEAN;
(*
 * liest die Variablen aus dem Config-File fileName
 * FALSE, wenn die Datei nicht geffnet werden konnte
 *)
(*
 * zuerst wird im angegeben Pfad gesucht, dann im
 * aktuellen Pfad, dann in dem Pfad, in dem die Programmdatei
 * liegt und zuletzt in $MAUSPATH.
 * Wenn das CFG dann immer noch nicht gefunden wurde,
 * liefern wir FALSE zurck.
 *)
VAR   result:        BOOLEAN;
      file:          File;
      filename,
      lRequestedPath,
      lCfgPath:      ARRAY [0 .. 127] OF CHAR;
      lCfgName:      ARRAY [0 .. 31] OF CHAR;
      path:          ARRAY [0 .. 255] OF CHAR;

PROCEDURE CheckVarname(varname: ARRAY OF CHAR): BOOLEAN;
(*
 * prfen, ob der Variablenname legal und noch nicht vergeben ist
 *)
VAR   result: BOOLEAN;
BEGIN
   result := TRUE;
   IF Find(varname)#NIL THEN
      IF debug THEN
         InOut.WriteString(varname); InOut.WriteLine(" schon bekannt");
      END;
      result := FALSE;
   END;
   RETURN result;
END CheckVarname;

PROCEDURE EnvExpand(VAR line: ARRAY OF CHAR);
(*
 * alle %VARNAME% in line durch den Wert der Environmentvariablen VARNAME
 * ersetzen
 *)
VAR   len,
      pos1,
      pos2:    CARDINAL;
      buffer:  ARRAY [0 .. LINELEN-1] OF CHAR;
      envvar:  ARRAY [0 .. ENVLEN-1] OF CHAR;
      value:   ARRAY [0 .. LINELEN-1] OF CHAR;
      quoted:  BOOLEAN;
BEGIN
   (*debug
   InOut.WriteString("vor dem Expandieren:  "); InOut.WriteLine(line);
   debug*)

   (*
    * Hm - nach MAUS-Sourcen werden Quotes beim expandieren nicht
    * beachtet; also kann ich doch meine ursprngliche, rekursive
    * Version benutzen  :)

   quoted := FALSE;
   len := Length(line);
   pos1 := 0;
   WHILE pos1<len DO
      IF NOT quoted THEN
         IF line[pos1]='%' THEN
            (*debug
            InOut.WriteLine("% gefunden");
            debug*)
            pos2 := pos1+1;
            WHILE (pos2<len) AND NOT quoted DO
               IF line[pos2]='%' THEN
                  Copy(line, pos1+1, pos2-pos1-1, envvar); (* Variablenname  *)
                  (*debug
                  InOut.WriteString("suche "); InOut.WriteLine(envvar);
                  debug*)
                  GetEnv(envvar, value);
                  (*debug
                  InOut.WriteString(">Wert "); InOut.WriteLine(value);
                  debug*)
                  Delete(line, pos1, pos2-pos1+1); (* Platzhalter lschen  *)
                  Insert(value, line, pos1);       (* Wert einsetzen       *)
                  INC(pos1, Length(value));
                  len := Length(line);
               ELSIF line[pos2]="'" THEN
                  quoted := TRUE;
               END;
               INC(pos2);
            END;
         ELSIF line[pos1]="'" THEN
            quoted := TRUE;
         END;
      ELSE
         (* so lange kein zweites Quotezeichen kommt,
          * bleiben wir im Quote-Modus
          *)
         quoted := line[pos1]#"'";
      END;
      INC(pos1);
   END;
   *)
   (* rekursive Version, ohne Beachtung von Quotes *)
   pos1 := Pos('%', line);                         (* erstes %             *)
   IF pos1<=HIGH(line) THEN
      Copy(line, pos1+1, LINELEN, buffer);
      pos2 := Pos('%', buffer);                    (* zweites %            *)
      IF pos2<=HIGH(buffer) THEN
         IF pos2>ENVLEN THEN
            Error(errEnvVarZuLang, line, lCfgName);
         ELSE
            Copy(buffer, 0, pos2, envvar);         (* Variablenname        *)
            (*debug
            InOut.WriteString("suche "); InOut.WriteLine(envvar);
            debug*)
            GetEnv(envvar, value);
            (*debug
            InOut.WriteString(">Wert "); InOut.WriteLine(value);
            debug*)
            Delete(line, pos1, pos2+2);            (* Platzhalter lschen  *)
            Insert(value, line, pos1);             (* Wert einsetzen       *)
            EnvExpand(line);                       (* weiter expandieren   *)
         END;
      END;
   END;
   (*debug
   InOut.WriteString("nach dem Expandieren: "); InOut.WriteLine(line);
   debug*)
END EnvExpand;

PROCEDURE AddVar(name, value: ARRAY OF CHAR): BOOLEAN;
VAR   result:  BOOLEAN;
      new,
      var:     Variable;
BEGIN
   (*debug
   InOut.WriteString(name); InOut.WriteString(" = ");
   InOut.WriteLine(value);
   debug*)
   result := FALSE;
   var := cfgVars;
   WHILE var^.next#NIL DO
      var := var^.next;
   END;
   NEW(var^.next);
   IF var^.next#NIL THEN
      new := var^.next;
      new^.next := NIL;
      new^.default.type := vUndef;
      new^.default.undefined := TRUE;
      new^.namelen := HIGH(name)+1;
      ALLOCATE(new^.name, new^.namelen);
      IF new^.name#NIL THEN
         Assign(name, new^.name^);
         new^.valuelen := HIGH(value)+1;
         ALLOCATE(new^.value, new^.valuelen);
         IF new^.value#NIL THEN
            Assign(value, new^.value^);
            result := TRUE;
         ELSE
            DEALLOCATE(new^.name, new^.namelen);
            DISPOSE(var^.next);
            var^.next := NIL;
         END;
      ELSE
         DISPOSE(var^.next);
         var^.next := NIL;
      END;
   END;
   IF debug THEN
      IF result THEN
         InOut.WriteString(" --> ");
         InOut.WriteString(name); InOut.WriteString(" = ");
         InOut.WriteLine(value);
      END;
   END;
   RETURN result;
END AddVar;

PROCEDURE Cut(VAR line: ARRAY OF CHAR);
(*
 * sucht den ersten ';', der nicht innerhalb von Quotes "'" steht
 * und schneidet den String dort ab (der ';' wird auch entfernt)
 *)
VAR   i,
      len:     CARDINAL;
      done,
      quoted:  BOOLEAN;
BEGIN
   (*debug
   InOut.WriteString("Cut: "); InOut.WriteLine(line);
   debug*)
   quoted := FALSE;
   done := FALSE;
   i := 0;
   len := Length(line);
   WHILE (NOT done)  AND (i<len) DO
      IF NOT quoted THEN
         IF line[i]=';' THEN
            line[i] := 0C;
            done := TRUE;
         ELSIF line[i]="'" THEN
            quoted := TRUE;
         END;
      ELSE
         (* so lange kein zweites Quotezeichen kommt,
          * bleiben wir im Quote-Modus
          *)
         quoted := line[i]#"'";
      END;
      INC(i);
   END;
   (*debug
   InOut.WriteString(">>>: "); InOut.WriteLine(line);
   debug*)
END Cut;

PROCEDURE ReadLines;
VAR   line:       ARRAY [0 .. LINELEN-1] OF CHAR;
      delimiter:  CARDINAL;
      varname:    ARRAY [0 .. VARLEN-1] OF CHAR;
BEGIN
   WHILE NOT EOF(file) DO
      ReadLine(file, line);
      IF debug THEN
         InOut.WriteLine(line);
      END;
      RemoveLeadingSpaces(line);
      (* Zeile nur bearbeiten, wenn nicht leer und kein Kommentar *)
      (*debug
      InOut.WriteLine(line);
      debug*)
      IF (line[0]#0C) AND (line[0]#';') THEN
         delimiter := Pos(':=', line);
         IF delimiter>HIGH(line) THEN
            IF debug THEN
               InOut.WriteLine("*** Zuweisung fehlt! ***");
            END;
            Error(errZuweisungFehlt, line, lCfgName);
         ELSIF delimiter>VARLEN THEN
            IF debug THEN
               InOut.WriteLine("*** Variablenname zu lang! ***");
            END;
            Error(errVarnameZuLang, line, lCfgName);
         ELSE
            Copy(line, 0, delimiter, varname);
            Trim(varname);
            Upper(varname);
            IF NOT CheckVarname(varname) THEN
               IF debug THEN
                  InOut.WriteLine("*** Variable bereits deklariert! ***");
               END;
               Error(errVarExists, varname, lCfgName);
            ELSE
               Delete(line, 0, delimiter+2);
               Trim(line);
               Cut(line);
               (*
               IF line[Length(line)-1]=';' THEN
                  line[Length(line)-1] := 0C;
                  RemoveTrailingSpaces(line);
               END;
               *)
               IF line[0]=0C THEN
                  IF debug THEN
                     InOut.WriteLine("*** leere Zuweisung! ***");
                  END;
                  Error(errLeereZuweisung, line, lCfgName);
               ELSE
                  (* endlich einen gltigen Wert ermittelt!
                   * jetzt noch ggf. die Werte von EnvVariablen einsetzen
                   * (Ersetzung von Steuerzeichen machen wir erst bei der Abfrage
                   * eines Strings, weil wir dazu evtl. Pascal-Strings brauchen)
                   *)
                  EnvExpand(line);
                  (* Sondervariablen 'ConfigFile' und 'ConfigInclude' bearbeiten *)
                  IF (CompareStr(varname, 'CONFIGFILE')=0) OR
                     (CompareStr(varname, 'CONFIGINCLUDE')=0) THEN
                     (* angegebenes Configfile lesen *)
                     RemoveQuotes(line, FALSE);
                     void.boolean := Read(line);
                  ELSE
                     (* Variable und Wert merken *)
                     IF NOT AddVar(varname, line) THEN
                        IF debug THEN
                           InOut.WriteLine("*** Ops! Speicher alle! ***");
                        END;
                        Error(errOutOfMem, '', lCfgName);
                     END;
                  END;
               END;
            END;
         END;
      ELSE
      (*debug
      InOut.WriteLine("--- leere Zeile oder Kommentar ---");
      debug*)
      END;
   END;
END ReadLines;

BEGIN
   result := FALSE;
   IF debug THEN
      InOut.WriteString("--- suche Config-File '");
      InOut.WriteString(fileName);
      InOut.WriteLine("' ---");
   END;
   Assign(fileName, filename);
   ExtractDirPath(fileName, lRequestedPath);
   ExtractFileName(fileName, lCfgName);
   IF debug THEN
      InOut.WriteString(" (1) suche im vorgegebenen Pfad: ");
      InOut.WriteLine(filename);
   END;
   IF Open(file, filename, aRead) THEN
      result := TRUE;
      Assign(requestedPath, lCfgPath);
   ELSE
      IF debug THEN
         InOut.WriteString(" (2) suche im aktuellen Pfad:    ");
         InOut.WriteLine(lCfgName);
      END;
      IF Open(file, lCfgName, aRead) THEN
         result := TRUE;
         GetCurDir(0, lCfgPath);
      ELSE
         GetEnv('MAUSPATH', filename);
         Append(filename, lCfgName);
         IF debug THEN
            InOut.WriteString(" (3) suche in MAUSPATH:          ");
            InOut.WriteLine(filename);
         END;
         IF Open(file, filename, aRead) THEN
            ExtractDirPath(filename, lCfgPath);
            result := TRUE;
         END;
      END;
      IF NOT result THEN
         Locate(lCfgName, "PATH", filename, result);
         IF result THEN
            IF debug THEN
               InOut.WriteString(" (4) suche in PATH:              ");
               InOut.WriteLine(filename);
            END;
            IF Open(file, filename, aRead) THEN
               ExtractDirPath(filename, lCfgPath);
            ELSE
               result := FALSE;
            END;
         END;
      END;
   END;
   CorrectPath(lCfgPath);
   IF result THEN
      IF NOT cfgOpen THEN
         cfgOpen := TRUE;
         Assign(lRequestedPath, requestedPath);
         Assign(lCfgPath, cfgPath);
         Assign(lCfgName, cfgName);
      END;
      IF debug THEN
         InOut.WriteString("--> gefunden und geffnet: ");
         InOut.WriteString(lCfgPath);
         InOut.WriteLine(lCfgName);
      END;
      ReadLines;
      Close(file);
      IF debug THEN
         InOut.WriteString("<-- geschlossen: ");
         InOut.WriteString(lCfgPath);
         InOut.WriteLine(lCfgName);
      END;
   ELSE
      IF debug THEN
         InOut.WriteString("--> NICHT gefunden!");
      END;
   END;
   RETURN result;
END Read;

PROCEDURE Dispose;
(*
 * gibt den durch Read belegten Speicher wieder frei
 * danach knnen keine Variablen mehr mit GetXXX abgefragt werden
 *)
VAR   var,
      next: Variable;
BEGIN
   IF cfgVars#NIL THEN
      var := cfgVars;
      next := var^.next;
      WHILE next#NIL DO
         var := next;
         IF var^.name#NIL THEN
            (*debug
            InOut.WriteString("Variable: "); InOut.WriteLine(var^.name^);
            InOut.WriteLine("DEALLOCATE(var^.name, var^.namelen);");
            debug*)
            DEALLOCATE(var^.name, var^.namelen);
         END;
         IF var^.value#NIL THEN
            (*debug
            InOut.WriteLine("DEALLOCATE(var^.value, var^.valuelen);");
            debug*)
            DEALLOCATE(var^.value, var^.valuelen);
         END;
         CASE var^.default.type OF
         | vString, vPath:
            IF var^.default.string#NIL THEN
               (*debug
               InOut.WriteLine("DEALLOCATE(var^.default.string, var^.default.size);");
               debug*)
               DEALLOCATE(var^.default.string, var^.default.size);
            END;
         ELSE
         END;
         next := var^.next;
         (*debug
         InOut.WriteLine("DEALLOCATE(var, SIZE(VarRec));");
         debug*)
         DISPOSE(var);
      END;
      DISPOSE(cfgVars);
      cfgVars := NIL;
   END;
END Dispose;


(* Die folgenden Prozeduren suchen eine Config-Variable aus dem eingelesenen
 * CFG und kopieren ihren Wert in die angegebene Modula-Variable.
 * Wurde die Variable im CFG nicht gefunden, oder ist deren Wert ungltig,
 * ist der Wert der Funktion FALSE. In diesem Fall gilt result:=default.
 * Konnte die Variable ohne Fehler gelesen werden, liefert die Funktion TRUE.
 *)

PROCEDURE GetReal(name: ARRAY OF CHAR; default: REAL;
                  VAR value: REAL): BOOLEAN;
(*
 * liefert einen REAL-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      ok:      BOOLEAN;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vReal) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vReal;
      var^.default.real := default;
      StringToReal(var^.value^, value, ok);
      IF NOT ok THEN
         Error(errNoReal, name, cfgName);
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetReal;

PROCEDURE GetInteger(name: ARRAY OF CHAR; default: INTEGER;
                     VAR value: INTEGER): BOOLEAN;
(*
 * liefert einen INTEGER-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      ok:      BOOLEAN;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vInteger) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vInteger;
      var^.default.integer := default;
      StringToInt(var^.value^, value, ok);
      IF NOT ok THEN
         Error(errNoInteger, name, cfgName);
      ELSE
         result := TRUE;
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetInteger;

PROCEDURE GetCardinal(name: ARRAY OF CHAR; default: CARDINAL;
                      VAR value: CARDINAL): BOOLEAN;
(*
 * liefert einen WORD-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      ok:      BOOLEAN;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vCardinal) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vCardinal;
      var^.default.cardinal := default;
      StringToCard(var^.value^, value, ok);
      IF NOT ok THEN
         Error(errNoCardinal, name, cfgName);
      ELSE
         result := TRUE;
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetCardinal;

PROCEDURE GetLongint(name: ARRAY OF CHAR; default: LONGINT;
                     VAR value: LONGINT): BOOLEAN;
(*
 * liefert einen LONGINT-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      ok:      BOOLEAN;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vLongint) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vLongint;
      var^.default.longint := default;
      StringToLongInt(var^.value^, value, ok);
      IF NOT ok THEN
         Error(errNoLongint, name, cfgName);
      ELSE
         result := TRUE;
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetLongint;

PROCEDURE GetBoolean(name: ARRAY OF CHAR; default: BOOLEAN;
                     VAR value: BOOLEAN): BOOLEAN;
(*
 * liefert einen BOOLEAN-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      buffer:  ARRAY [0 .. 4] OF CHAR;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vBoolean) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vBoolean;
      var^.default.boolean := default;
      Copy(var^.value^, 0, 5, buffer);
      Upper(buffer);
      IF CompareStr(buffer, 'TRUE')=0 THEN
         value := TRUE;
         result := TRUE;
      ELSIF CompareStr(buffer, 'FALSE')=0 THEN
         value := FALSE;
         result := TRUE;
      ELSE
         Error(errNoBoolean, name, cfgName);
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetBoolean;

PROCEDURE GetChar(name: ARRAY OF CHAR; default: CHAR;
                  VAR value: CHAR): BOOLEAN;
(*
 * liefert einen CHAR-Parameter zurck
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
BEGIN
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vChar) THEN
         Error(errWrongType, name, cfgName);
      END;
      var^.default.type := vChar;
      var^.default.char := default;
      IF Length(var^.value^)=1 THEN
         value := var^.value^[0];
         result := TRUE;
      ELSE
         Error(errNoChar, name, cfgName);
      END;
   ELSE
      value := default;
   END;
   RETURN result;
END GetChar;

PROCEDURE EscExpand(VAR line: ARRAY OF CHAR;
                    VAR null: BOOLEAN);
(*
 * Escapes (#) ersetzen
 *)

VAR   len,
      i, j,
      pos:     CARDINAL;
      num:     ARRAY [0 .. 2] OF CHAR;
      quoted:  BOOLEAN;
      val:     CARDINAL;
      ok:      BOOLEAN;
BEGIN
   (*debug
   InOut.WriteString("mit '#':  "); InOut.WriteLine(line);
   debug*)
   null := FALSE;
   quoted := FALSE;
   len := Length(line);
   pos := 0;
   WHILE pos<len DO
      IF NOT quoted THEN
         IF line[pos]='#' THEN
            (*debug
            InOut.WriteLine("# gefunden");
            debug*)
            IF line[pos+1]='#' THEN       (* ## => # *)
               INC(pos);
               FOR i := pos TO len-1 DO
                  line[i] := line[i+1];
               END;
               DEC(len);
            ELSE
               i := pos+1;
               WHILE (line[i]>='0') AND (line[i]<='9') DO
                  INC(i);
               END;
               DEC(i, pos+1);
               IF i>3 THEN i := 3 END;
               Copy(line, pos+1, i, num);
               (*debug
               InOut.WriteString("umwandeln:  "); InOut.WriteLine(num);
               debug*)
               StringToCard(num, val, ok);
               IF (NOT ok) OR (val>255) THEN
                  (* Error *)
                  (*debug
                  InOut.WriteLine(" :-( nicht geklappt ");
                  debug*)
               ELSE
                  null := (val=0);
                  line[pos] := CHR(val);
                  FOR j := pos+1 TO len-1-i DO
                     line[j] := line[j+i];
                  END;
                  DEC(len, i);
               END;
            END;
         ELSIF line[pos]="'" THEN
            quoted := TRUE;
         END;
      ELSE
         (* so lange kein zweites Quotezeichen kommt,
          * bleiben wir im Quote-Modus
          *)
         quoted := line[pos]#"'";
      END;
      INC(pos);
   END;
   IF null THEN
      FOR i := len TO 1 BY -1 DO
         line[i] := line[i-1];
      END;
      line[0] := CHR(len);
   END;
   line[len] := 0C;
   (*debug
   InOut.WriteString("ohne '#': "); InOut.WriteLine(line);
   debug*)
END EscExpand;


PROCEDURE GetString(name: ARRAY OF CHAR; default: ARRAY OF CHAR;
                    VAR value: ARRAY OF CHAR;
                    VAR null:  BOOLEAN): BOOLEAN;
(*
 * liefert einen STRING-Parameter zurck
 *
 * Falls null=TRUE, enthlt value einen Null-Character und ORD(value[0])
 * gibt die Anzahl der noch folgenden gltigen Zeichen an!
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
BEGIN
   result := FALSE;
   null := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      (*debug
      InOut.WriteString("** gesucht:  "); InOut.WriteLine(name);
      InOut.WriteString("** gefunden: "); InOut.WriteLine(var^.name^);
      debug*)
      IF (var^.default.type#vUndef) AND (var^.default.type#vString) THEN
         Error(errWrongType, name, cfgName);
      END;
      IF (var^.default.type=vString) THEN
         IF var^.default.string#NIL THEN
            DEALLOCATE(var^.default.string, var^.default.size);
         END;
      END;
      var^.default.type := vString;
      (* Platz fr String allozieren *)
      var^.default.size := Length(default)+1;
      ALLOCATE(var^.default.string, var^.default.size);
      IF var^.default.string#NIL THEN
         Assign(default, var^.default.string^);
      ELSE
         Error(errOutOfMem, name, cfgName);
         var^.default.size := 0;
      END;
      Assign(var^.value^, value);
      result := TRUE;
   ELSE
      Assign(default, value);
   END;
   (* Escapes (#123) umwandeln! *)
   EscExpand(value, null);
   RemoveQuotes(value, null);
   RETURN result;
END GetString;

PROCEDURE GetPath(name: ARRAY OF CHAR; default: ARRAY OF CHAR;
                  VAR value: ARRAY OF CHAR): BOOLEAN;
(*
 * liefert einen Directory-Pfad-Parameter zurck ('\' am Ende);
 *)
VAR   result:  BOOLEAN;
      var:     Variable;
      null:    BOOLEAN;
BEGIN
   (*debug
   InOut.WriteString("suche nach: "); InOut.WriteLine(name);
   debug*)
   result := FALSE;
   Upper(name);
   var := Find(name);
   IF var#NIL THEN
      IF (var^.default.type#vUndef) AND (var^.default.type#vPath) THEN
         Error(errWrongType, name, cfgName);
      END;
      IF (var^.default.type=vPath) THEN
         IF var^.default.string#NIL THEN
            DEALLOCATE(var^.default.string, var^.default.size);
         END;
      END;
      var^.default.type := vPath;
      (* Platz fr String allozieren *)
      var^.default.size := Length(default)+1;
      (*debug
      InOut.WriteString("size: "); InOut.WriteCard(var^.default.size, 0);
      InOut.WriteLn;
      debug*)
      ALLOCATE(var^.default.string, var^.default.size);
      IF var^.default.string#NIL THEN
         (*debug
         InOut.WriteString("default: "); InOut.WriteCard(Length(default), 0);
         InOut.WriteLn;
         debug*)
         Assign(default, var^.default.string^);
      ELSE
         Error(errOutOfMem, name, cfgName);
         var^.default.size := 0;
      END;
      Assign(var^.value^, value);
      (*debug
      InOut.WriteString("> gefunden: "); InOut.WriteLine(value);
      debug*)
      result := TRUE;
   ELSE
      (*debug
      InOut.WriteLine("> nicht gefunden");
      debug*)
      Assign(default, value);
   END;
   EscExpand(value, null);
   IF NOT null THEN
      RemoveQuotes(value, FALSE);
   ELSE
      result := FALSE;
   END;
   RETURN result;
END GetPath;

PROCEDURE SetDebugging(on: BOOLEAN);
BEGIN
   debug := on;
END SetDebugging;

BEGIN
   ModReg.Register('ConfigFile',0,0,' ',1996,10,16,Dispose);
   debug := FALSE;
   (* es ist noch kein CFG-File geffnet *)
   cfgOpen := FALSE;
   NEW(cfgVars);
   WITH cfgVars^ DO
      default.type := vUndef;
      name := NIL;
      value := NIL;
      default.undefined := TRUE;
      next := NIL;
   END;

END ConfigFile.

