IMPLEMENTATION MODULE Dictionaries;

(*
 * In einem Dictionary knnen Paare (key, data) gespeichert werden.
 * Der Zugriffe erfolgt ber key. Es knnen nicht mehrere Paare mit dem
 * gleichen key in einem Dictionary sein.
 *)


FROM     SYSTEM         IMPORT   ADDRESS;
FROM     Storage        IMPORT   ALLOCATE, DEALLOCATE;
IMPORT   ModReg;


TYPE     Entry          =  POINTER TO EntryRec;

         EntryRec       =  RECORD
                              key,
                              data: ADDRESS;
                              next: Entry;
                           END;

         Dictionary     =  POINTER TO RECORD
                              first,
                              last:    Entry;
                              entries: LONGCARD;
                           END;


PROCEDURE HasKey(aEntry, key: ADDRESS): BOOLEAN;
VAR   entry:   Entry;
BEGIN
   entry := aEntry;
   RETURN (entry^.key=key);
END HasKey;



PROCEDURE Create(VAR dict:  Dictionary): BOOLEAN;
BEGIN
   ALLOCATE(dict, SIZE(dict^));
   IF dict#NIL THEN
      dict^.first := NIL;
      dict^.last := NIL;
   END;
   dict^.entries := 0L;
   RETURN (dict#NIL);
END Create;

PROCEDURE Delete(VAR dict:  Dictionary): BOOLEAN;
VAR   entry,
      next:    Entry;
BEGIN
   next := dict^.first;
   WHILE next#NIL DO
      entry := next;
      next := entry^.next;
      DEALLOCATE(entry, SIZE(entry^));
   END;
   DEALLOCATE(dict, SIZE(dict^));
   RETURN TRUE;
END Delete;


PROCEDURE Empty(dict:   Dictionary): BOOLEAN;
BEGIN
   RETURN (dict^.entries=0L);
END Empty;

PROCEDURE Entries(dict: Dictionary): LONGCARD;
BEGIN
   RETURN dict^.entries;
END Entries;


PROCEDURE Set(dict: Dictionary;
              key,
              data: ADDRESS): BOOLEAN;
   (*
    * FALSE, falle key nicht eindeutig ist.
    *)
VAR   result:  BOOLEAN;
      index:   LONGCARD;
      entry:   Entry;
      olddata: ADDRESS;
BEGIN
   result := FALSE;
   IF NOT Get(dict, key, olddata) THEN
      result := TRUE;
      ALLOCATE(entry, SIZE(entry^));
      IF entry#NIL THEN
         entry^.key := key;
         entry^.data := data;
         entry^.next := NIL;
         INC(dict^.entries);
      END;
      IF dict^.first=NIL THEN
         dict^.first := entry;
         dict^.last := entry;
      ELSE
         dict^.last^.next := entry;
      END;
   END;
   RETURN result;
END Set;


PROCEDURE Get(    dict: Dictionary;
                  key:  ADDRESS;
              VAR data: ADDRESS): BOOLEAN;
   (*
    * TRUE, falls key gefunden wurde,
    * FALSE sonst. Bei Mierfolg erhlt data den Wert NIL.
    *)
VAR   result:  BOOLEAN;
      next,
      entry:   Entry;
BEGIN
   data := NIL;
   result := FALSE;
   IF dict^.entries>0L THEN
      entry := dict^.first;
      WHILE (entry^.key#key) AND (entry^.next#NIL) DO
         entry := entry^.next;
      END;
      IF entry^.key=key THEN
         result := TRUE;
         data := entry^.data;
      END;
   END;
   RETURN result;
END Get;

PROCEDURE Remove(dict: Dictionary;
                 key:  ADDRESS): BOOLEAN;
   (*
    * FALSE, falls key nicht gefunden wurde,
    * TRUE sonst. Im Erfolgsfall, wird das Paar (key, data) aus
    * dict entfernt.
    *)
VAR   result:  BOOLEAN;
      prev,
      entry:   Entry;
BEGIN
   result := FALSE;
   IF dict^.entries=1L THEN
      entry := dict^.first;
      IF entry^.key=key THEN
         DEALLOCATE(entry, SIZE(entry^));
         dict^.first := NIL;
         dict^.last := NIL;
         DEC(dict^.entries);
      END;
   ELSIF dict^.entries>1L THEN
      entry := dict^.first;
      prev := entry;
      WHILE (entry^.key#key) AND (entry^.next#NIL) DO
         prev := entry;
         entry := entry^.next;
      END;
      IF entry^.key=key THEN
         result := TRUE;
         IF entry=dict^.first THEN
            dict^.first := entry^.next;
         ELSIF entry=dict^.last THEN
            prev^.next := entry^.next;
            dict^.last := prev;
         ELSE
            prev^.next := entry^.next;
         END;
         DEALLOCATE(entry, SIZE(entry^));
         DEC(dict^.entries);
      END;
   END;
   RETURN result;
END Remove;


PROCEDURE Scan(dict: Dictionary; proc: ScanProc);
   (*
    * ruft fr jeden Eintrag des Dictionaries proc auf
    *)
VAR   i:       LONGCARD;
      next,
      entry:   Entry;
BEGIN
   next := dict^.first;
   WHILE next#NIL DO
      entry := next;
      next := entry^.next;
      proc(entry^.data);
   END;
END Scan;


BEGIN
   ModReg.Register('Dictionaries',0,0,' ',1996,10,16,PROC(NIL));

END Dictionaries.
