IMPLEMENTATION MODULE FileIO;

(*
 * 29.09.1996  -  Bugfix in ReadBytes: Wenn die zu lesende Anzahl Bytes
 *                kleiner war als der Filebuffer, blieb ReadBytes in einer
 *                Endlosschleife hngen
 *             -  Bugfix in ReadBytes: done wurde nicht gesetzt
 * 13.10.1996  -  TermProcedure wird nicht mehr benutzt. Statt dessen
 *                mu am Programmende FileIO.Exit aufgerufen werden
 *             -  statt Files.CreateTmp wird jetzt eine eigene Routine
 *                benutzt
 * 16.10.1996  -  CloseAll neu implementiert
 *             -  Exit wird von TermProcedure automatisch aufgerufen
 *                (mehrfacher Aufruf schadet nicht)
 *             -  CloseAll kann gefahrlos auch nach Ausfhrung von Exit
 *                aufgerufen werden
 *             -  benutzt ModReg.Register
 * 18.08.1997  -  auskommentierten alten Code entfernt
 *             -  in AddFile wird geprft, ob ALLOCATE geglckt ist
 *)

FROM     SYSTEM            IMPORT   ADDRESS, ADR;
FROM     System            IMPORT   Move, TermProcedure;
FROM     Storage           IMPORT   ALLOCATE, DEALLOCATE;
IMPORT   Files;
FROM     Strings           IMPORT   Append, Length, CompareStr;
FROM     Strings1          IMPORT   Assign, CorrectPath;
FROM     ASCII             IMPORT   CR, LF, CtrlZ;
FROM     RandomGen         IMPORT   RandomCard;
IMPORT   ModReg;
IMPORT   void;
IMPORT   Dictionaries;
(*
IMPORT InOut;
*)

CONST    MAXTRIES    =  32;   (* Versuche ein temporres File anzulegen *)

TYPE     File        =  POINTER TO FileRec;
         BufferState =  (bOff, bRead, bWrite);
         FileRec     =  RECORD
                           id:      INTEGER;
                           state:   INTEGER;
                           mode:    BITSET;
                              (* physikalische Posistion des Schreib-
                               * Lesezeigers. Diese entspricht dem Ende
                               * des Lesepuffers *)
                           physPos,
                              (* logische Position des Schreib-Lese-
                               * Zeigers. Diese Position ist von auen
                               * sichtbar *)
                           logPos,
                           length:  LONGCARD;
                           buffer:  RECORD
                                       state:   BufferState;
                                          (* Startadresse des Puffers   *)
                                       adr:     ADDRESS;
                                          (* Gre des Puffers          *)
                                       size:    CARDINAL;
                                          (* Offset des Pufferanfangs   *)
                                          (* innerhalb der Datei        *)
                                       start:   LONGCARD;
                                          (* gltige Bytes im Puffer    *)
                                       full:    CARDINAL;
                                    END;
                        END;

         PtrChar     =  POINTER TO CHAR;
         PtrString   =  POINTER TO ARRAY [0 .. 32767] OF CHAR;

VAR      files:      Dictionaries.Dictionary;

         CRLF:       ARRAY [0 .. 1] OF CHAR;


PROCEDURE AddFile(id:      INTEGER;
                  state:   INTEGER;
                  mode:    BITSET): File;
VAR   result:  File;
BEGIN
   ALLOCATE(result, SIZE(FileRec));
   IF result#NIL THEN
      IF NOT Dictionaries.Set(files, result, result) THEN
         DEALLOCATE(result, SIZE(FileRec));
         result := NIL;
      ELSE
         result^.id := id;
         result^.state := state;
         result^.mode := mode;
         result^.physPos := 0L;
         result^.logPos := 0L;
         result^.length := 0L;
         ALLOCATE(result^.buffer.adr, defaultBufferSize);
         IF result^.buffer.adr#NIL THEN
            result^.buffer.size := defaultBufferSize;
         ELSE
            result^.buffer.size := 0;
         END;
         result^.buffer.state := bOff;
         result^.buffer.full := 0;
      END;
   END;
   RETURN result;
END AddFile;

PROCEDURE RemoveFile(file: File);
VAR   prev: File;
      done: BOOLEAN;
BEGIN
   IF file^.buffer.size>0 THEN
      DEALLOCATE(file^.buffer.adr, file^.buffer.size);
   END;
   IF Dictionaries.Remove(files, file) THEN
      DEALLOCATE(file, SIZE(FileRec));
   END;
END RemoveFile;

PROCEDURE BufferSize(file: File;
                     set:  CARDINAL): CARDINAL;
(*
 * Wert der Funktion ist die Gre des Lesepuffers fr file in Bytes.
 * Falls set>0, wird versucht, den Lesepuffer auf die angegebene Gre
 * zu setzen.
 * Fr set=0 wird lediglich die aktuelle Gre zurckgegeben.
 * Wird ein File neu geffnet, werden defaultBufferSize Bytes als Puffer
 * reserviert.
 *)
BEGIN
   IF set>0 THEN
      DEALLOCATE(file^.buffer.adr, file^.buffer.size);
      ALLOCATE(file^.buffer.adr, set);
      IF file^.buffer.adr#NIL THEN
         file^.buffer.size := set;
      ELSE
         file^.buffer.size := 0;
      END;
      file^.buffer.state := bOff;
      file^.buffer.full := 0;
   END;
   RETURN file^.buffer.size;
END BufferSize;

PROCEDURE ClearBuffer(file: File; mode: BufferState);
VAR   written: CARDINAL;
BEGIN
   IF file^.buffer.state=bWrite THEN
      Files.Seek(file^.id, Files.SEEKABS, file^.buffer.start);
      Files.Write(file^.id, file^.buffer.adr, file^.buffer.full, written);
      IF written#file^.buffer.size THEN
         file^.state := -99;
      END;
      file^.physPos := file^.buffer.start+LONG(written);
   END;
   IF file^.logPos#file^.physPos THEN
      Files.Seek(file^.id, Files.SEEKABS, file^.logPos);
      file^.physPos := file^.logPos;
   END;
   file^.buffer.state := mode;
   file^.buffer.start := file^.physPos;
   file^.buffer.full := 0;
END ClearBuffer;

PROCEDURE GoLogPos(file: File);
BEGIN
   IF file^.logPos#file^.physPos THEN
      ClearBuffer(file, bOff);
      Files.Seek(file^.id, Files.SEEKABS, file^.logPos);
      file^.physPos := file^.logPos;
   END;
END GoLogPos;

PROCEDURE GetFileDesc(file:   File): INTEGER;
(*
 * liefert den DOS-Filedescriptor fr file
 * braucht man z.B. um die Funktionen in Files nutzen zu knnen.
 * -1, falls file illegal!
 *)
BEGIN
   IF file#NIL THEN
      RETURN file^.id;
   ELSE
      RETURN NOFILE;
   END;
END GetFileDesc;

PROCEDURE Create(VAR file:       File;
                     name:       ARRAY OF CHAR;
                     accessMode: AccessMode;
                     createMode: CreateMode): BOOLEAN;
(*
 * file neu anlegen.
 *
 * createMode        Verhalten
 *  cNew              neu anlegen, oder Abbruch falls bereits vorhanden
 *  cAppend           neu anlegen, oder ffnen falls bereits vorhanden
 *  cClobber          neu anlegen; berschreiben falls bereits vorhanden
 *
 * TRUE genau dann, wenn die Datei angelegt bzw. geffnet werden konnte
 *)
VAR   result:  BOOLEAN;
      fMode:   BITSET;
      id:      INTEGER;
      length:  LONGCARD;
BEGIN
   result := FALSE;
   file := NIL;
   length := 0L;
   CASE accessMode OF
   | aRead:    fMode := Files.READ;
   | aWrite:   fMode := Files.WRITE;
   | aRW:      fMode := Files.IO;
   END;
   CASE createMode OF
   | cNew:     Files.CreateNew(id, name, fMode);
   | cAppend:  Files.Open(id, name, fMode);
               IF id<0 THEN
                  Files.Create(id, name, fMode)
               ELSE
                  Files.Seek(id, Files.SEEKEND, length);
               END;
   | cClobber: Files.Create(id, name, fMode);
   END;
   IF id>=0 THEN
      file := AddFile(id, Files.FileStatus, fMode);
      IF file#NIL THEN
         file^.length := length;
         file^.physPos := length;
         file^.logPos := length;
         result := (Files.FileStatus=0);
      ELSE
         Files.Close(id);
      END;
   END;
   RETURN result;
END Create;

PROCEDURE CreateTmp(VAR file:       File;
                    VAR name:       ARRAY OF CHAR;
                        mode:       AccessMode): BOOLEAN;
(*
 * temporres File neu anlegen
 * name sollte den Pfad enthalten; nach dem Aufruf steht dort der komplette
 * Filename
 *)
VAR   result:  BOOLEAN;
      fMode:   BITSET;
      i,
      id:      INTEGER;
      tries:   CARDINAL;
      fname:   ARRAY [0 .. 12] OF CHAR;
      path:    PathName;
BEGIN
   result := FALSE;
   file := NIL;
   CASE mode OF
   | aRead:    fMode := Files.READ;
   | aWrite:   fMode := Files.WRITE;
   | aRW:      fMode := Files.IO;
   END;
   (* ersetzt Files.CreateTmp *)
   tries := 0;
   Assign(name, path);
   CorrectPath(path);
   REPEAT
      FOR i := 0 TO 7 DO
         fname[i] := CHR(RandomCard(ORD("A"), ORD("Z")));
      END;
      fname[8] := 0C;
      Assign(path, name);
      Append(name, fname);
      INC(tries);
   UNTIL Create(file, name, mode, cNew) OR (tries>MAXTRIES);
   result := (tries<=MAXTRIES);
   RETURN result;
END CreateTmp;

PROCEDURE Open(VAR file:   File;
                   name:   ARRAY OF CHAR;
                   mode:   AccessMode): BOOLEAN;
(*
 * file ffnen
 * FALSE, falls ein Fehler aufgetreten ist
 *)
VAR   result:  BOOLEAN;
      fMode:   BITSET;
      id:      INTEGER;
      pos:     LONGCARD;
BEGIN
   result := FALSE;
   file := NIL;
   CASE mode OF
   | aRead:    fMode := Files.READ;
   | aWrite:   fMode := Files.WRITE;
   | aRW:      fMode := Files.IO;
   END;
   IF CompareStr(name, STDIN)=0 THEN
      id := 0;
   ELSIF CompareStr(name, STDOUT)=0 THEN
      id := 1;
   ELSIF CompareStr(name, STDERR)=0 THEN
      id := 2;
   ELSIF CompareStr(name, STDCOM)=0 THEN
      id := 3;
   ELSIF CompareStr(name, STDPRN)=0 THEN
      id := 4;
   ELSE
      Files.Open(id, name, fMode);
   END;
   IF id>=0 THEN
      file := AddFile(id, Files.FileStatus, fMode);
      IF file#NIL THEN
         Files.Seek(id, Files.SEEKEND, file^.length);
         pos := 0L;
         Files.Seek(id, Files.SEEKABS, pos);
         result := (Files.FileStatus=0);
      ELSE
         Files.Close(id);
      END;
   END;
   RETURN result;
END Open;

PROCEDURE Close(file: File);
(*
 * file schlieen
 *)
BEGIN
   IF file#NIL THEN
      ClearBuffer(file, bOff);
      Files.Close(file^.id);
      RemoveFile(file);
   END;
END Close;

PROCEDURE DisposeFile(file: File);
(*
 * nur den Speicher freigeben, aber nicht aus dem Dictionary lschen
 * (fr CloseAll)
 *)
BEGIN
   IF file#NIL THEN
      ClearBuffer(file, bOff);
      Files.Close(file^.id);
      IF file^.buffer.size>0 THEN
         DEALLOCATE(file^.buffer.adr, file^.buffer.size);
      END;
      DEALLOCATE(file, SIZE(FileRec));
   END;
END DisposeFile;

PROCEDURE CloseAll;
(*
 * schliet alle ber FileIO.Open oder FileIO.Create geffneten Dateien
 *)
BEGIN
   IF files#NIL THEN
      Dictionaries.Scan(files, DisposeFile);
      void.boolean := Dictionaries.Delete(files);
      void.boolean := Dictionaries.Create(files);
   END;
END CloseAll;

PROCEDURE Flush(file: File);
(*
 * schreibt den Pufferinhalt auf Platte
 *)
BEGIN
   IF file#NIL THEN
      ClearBuffer(file, bOff);
   END;
END Flush;

PROCEDURE Seek(file:    File;
               mode:    SeekMode;
               offset:  LONGCARD): LONGCARD;
(*
 * auf offset positionieren
 * Wert der Funktion ist der tatschliche Offset
 *)
VAR   result:  LONGCARD;
      sMode:   Files.SeekMode;
BEGIN
   CASE mode OF
   | sBegin:
      IF offset<file^.length THEN
         result := offset;
      ELSE
         result := file^.length;
      END;
   | sCurrent:
      result := file^.logPos+offset;
      IF result>file^.length THEN
         result := file^.length;
      END;
   | sEnd:
      IF offset<file^.length THEN
         result := file^.length-offset;
      ELSE
         result := 0L;
      END;
   END;
   file^.logPos := result;
   RETURN result;
END Seek;

PROCEDURE Seek0(file: File): LONGCARD;
(*
 * liefert die aktuelle Position des File-Pointers
 *)
BEGIN
   RETURN file^.logPos;
END Seek0;

PROCEDURE State(file:   File): INTEGER;
BEGIN
   IF file#NIL THEN
      RETURN file^.state;
   ELSE
      RETURN NOFILE;
   END;
END State;

PROCEDURE EOF(file:  File): BOOLEAN;
(*
 * TRUE, falls das Fileende erreicht ist
 *)
VAR   offset:  LONGCARD;
      buffer:  CHAR;
      read:    CARDINAL;
BEGIN
   RETURN (file^.logPos >= file^.length);
END EOF;

PROCEDURE FillBuffer(file: File);
BEGIN
   (*debug
   InOut.WriteLine("> FillBuffer");
   InOut.WriteString(" Position: ");
   InOut.WriteLongCard(file^.physPos, 6);
   InOut.WriteLn;
   debug*)
   IF (file^.physPos>file^.logPos) OR
      ((file^.physPos+LONG(file^.buffer.size))<file^.logPos) THEN
      GoLogPos(file);
   END;
   file^.buffer.start := file^.physPos;
   Files.Read(file^.id, file^.buffer.adr, file^.buffer.size, file^.buffer.full);
   file^.state := Files.FileStatus;
   INC(file^.physPos, file^.buffer.full);
   (*debug
   InOut.WriteLine("< FillBuffer");
   debug*)
END FillBuffer;

PROCEDURE CheckBuffer(file:   File;
                      mode:   BufferState);
BEGIN
   CASE mode OF
   | bRead:
      IF file^.buffer.state#bWrite THEN
         IF (file^.buffer.full=0) OR
            (file^.logPos<file^.buffer.start) OR
            (file^.logPos>(file^.buffer.start+LONG(file^.buffer.full))) THEN
            FillBuffer(file);
         END;
      ELSE  (* Write *)
         ClearBuffer(file, mode);
      END;
   | bWrite:
      IF file^.buffer.state=bWrite THEN
         IF file^.logPos#(file^.buffer.start+LONG(file^.buffer.full)) THEN
            ClearBuffer(file, mode);
         END;
      ELSE
         ClearBuffer(file, mode);
      END;
   ELSE
      HALT;
   END;
END CheckBuffer;

PROCEDURE Read(file:    File): CHAR;
(*
 * ein Zeichen aus file lesen
 *)
VAR   result:  CHAR;
      adr:     PtrChar;
BEGIN
   IF file^.logPos < file^.length THEN
      CheckBuffer(file, bRead);
      adr := ADDRESS(file^.logPos-file^.buffer.start
                     +LONGCARD(file^.buffer.adr));
      result := adr^;
      INC(file^.logPos);
      IF file^.logPos>(file^.buffer.start+LONG(file^.buffer.full)) THEN
         file^.buffer.full := 0;
      END;
   ELSE
      result := CtrlZ;
   END;
   RETURN result;
END Read;

PROCEDURE ReadLine(    file:  File;
                   VAR str:   ARRAY OF CHAR);
(*
 * eine Zeile aus file lesen
 *)
VAR   done:    BOOLEAN;
      start:   PtrString;
      strPos,
      bufPos,
      strLen,
      bufLen:  CARDINAL;
BEGIN
   (*debug
   InOut.WriteLine("> ReadLine");
   debug*)
   CheckBuffer(file, bRead);
   done := FALSE;
   strLen := HIGH(str);
   strPos := 0;
   REPEAT
      start := ADDRESS(file^.logPos-file^.buffer.start
                       +LONGCARD(file^.buffer.adr));
      bufLen := SHORT(file^.buffer.start+LONG(file^.buffer.full)-file^.logPos);
      bufPos := 0;
      WHILE NOT done AND (bufLen>0) AND (strLen>0) DO
         CASE start^[bufPos] OF
           CR:
            done := TRUE;
            str[strPos] := 0C;
            INC(bufPos);
            INC(file^.logPos);
            DEC(bufLen);
            IF bufLen=0 THEN
               FillBuffer(file);
               start := ADDRESS(file^.logPos-file^.buffer.start
                                +LONGCARD(file^.buffer.adr));
               bufLen := SHORT(file^.buffer.start+LONG(file^.buffer.full)-file^.logPos);
               bufPos := 0;
            END;
            IF start^[bufPos]=LF THEN
               INC(file^.logPos);
            END;
         | LF:
            done := TRUE;
            str[strPos] := 0C;
            INC(bufPos);
            INC(file^.logPos);
            DEC(bufLen);
            IF bufLen=0 THEN
               FillBuffer(file);
               start := ADDRESS(file^.logPos-file^.buffer.start
                                +LONGCARD(file^.buffer.adr));
               bufLen := SHORT(file^.buffer.start+LONG(file^.buffer.full)-file^.logPos);
               bufPos := 0;
            END;
            IF start^[bufPos]=CR THEN
               INC(file^.logPos);
            END;
         | CtrlZ:
            done := TRUE;
            INC(bufPos);
            INC(file^.logPos);
            DEC(bufLen);
         ELSE
            str[strPos] := start^[bufPos];
            INC(bufPos);
            INC(strPos);
            INC(file^.logPos);
            DEC(bufLen);
            DEC(strLen);
         END;
      END;
      IF strPos<strLen THEN
         str[strPos] := 0C;
      END;
      IF bufLen=0 THEN
         FillBuffer(file);
      END;
   UNTIL done OR (strLen=0) OR EOF(file);
   (*debug
   InOut.WriteLine("< ReadLine");
   debug*)
END ReadLine;

PROCEDURE ReadBytes(file:  File;
                    adr:   ADDRESS;
                    len:   CARDINAL;
                    VAR done:  CARDINAL);
(*
 * len Bytes aus file nach adr lesen
 * done gibt an, wieviele Bytes tatschlich gelesen wurden
 *)
VAR   start:   ADDRESS;
      bufLen:  CARDINAL;
BEGIN
   done := 0;
   CheckBuffer(file, bRead);
   WHILE len>0 DO
      start := ADDRESS(file^.logPos-file^.buffer.start
                       +LONGCARD(file^.buffer.adr));
      bufLen := SHORT(file^.buffer.start+LONG(file^.buffer.full)-file^.logPos);
      IF len <= bufLen THEN
         Move(start, adr, len);
         INC(file^.logPos, len);
         INC(done, len);
         len := 0;
      ELSE
         Move(start, adr, bufLen);
         INC(file^.logPos, bufLen);
         FillBuffer(file);
         DEC(len, bufLen);
         INC(adr, bufLen);
         INC(done, bufLen);
      END;
   END;  (* WHILE *)
   IF file^.logPos>(file^.buffer.start+LONG(file^.buffer.full)) THEN
      file^.buffer.full := 0;
   END;
END ReadBytes;

PROCEDURE Write(file:   File;
                char:   CHAR);
(*
 * char nach file schreiben
 *)
VAR   adr:   ADDRESS;
BEGIN
   CheckBuffer(file, bWrite);
   IF file^.buffer.full=file^.buffer.size THEN
      ClearBuffer(file, bWrite);
   END;
   adr := ADDRESS(LONGCARD(file^.buffer.adr)+LONG(file^.buffer.full));
   Move(ADR(char), adr, 1);
   INC(file^.buffer.full);
   INC(file^.logPos);
   IF (file^.logPos) > file^.length THEN
      file^.length := file^.logPos;
   END;
END Write;

PROCEDURE WriteBytes(    file: File;
                         adr:  ADDRESS;
                         len:  CARDINAL;
                     VAR done: CARDINAL);
(*
 * len Bytes von adr in file schreiben
 * done gibt an, wieviele Bytes tatschlich geschrieben wurden
 *)
VAR   from, to:   ADDRESS;
      left,
      bytes:      CARDINAL;
BEGIN
   CheckBuffer(file, bWrite);
   from := adr;
   left := len;
   done := 0;
   WHILE left > 0 DO
      IF file^.buffer.full=file^.buffer.size THEN
         ClearBuffer(file, bWrite);
      END;
      to := ADDRESS(LONGCARD(file^.buffer.adr)+LONG(file^.buffer.full));
      bytes := (file^.buffer.size-file^.buffer.full);
      IF left<bytes THEN bytes := left END;
      Move(from, to, bytes);
      DEC(left, bytes);
      INC(from, bytes);
      INC(file^.buffer.full, bytes);
      INC(file^.logPos, bytes);
      INC(done, bytes);
   END;
   IF (file^.logPos) > file^.length THEN
      file^.length := file^.logPos;
   END;
END WriteBytes;

PROCEDURE WriteString(file:   File;
                      str:    ARRAY OF CHAR);
(*
 * str mach file schreiben
 *)
VAR   from, to:   ADDRESS;
      left,
      bytes:      CARDINAL;
BEGIN
   WriteBytes(file, ADR(str), Length(str), void.cardinal);
END WriteString;

PROCEDURE WriteLn(file: File);
(*
 * CR,LF nach file schreiben
 *)
BEGIN
   WriteBytes(file, ADR(CRLF), 2, void.cardinal);
END WriteLn;

PROCEDURE WriteLine(file:  File;
                    str:   ARRAY OF CHAR);
(*
 * str+CR,LF nach file schreiben
 *)
BEGIN
   WriteBytes(file, ADR(str), Length(str), void.cardinal);
   WriteBytes(file, ADR(CRLF), 2, void.cardinal);
END WriteLine;

PROCEDURE Exit;
(*
 * am Programmende aufrufen, um globalen Speicher freizugeben
 *)
BEGIN
   IF files#NIL THEN
      CloseAll;
      void.boolean := Dictionaries.Delete(files);
      files := NIL;
   END;
END Exit;


BEGIN
   files := NIL;
   void.boolean := Dictionaries.Create(files);

   ModReg.Register('FileIO',0,0,' ', 1997,08,18, Exit);
   
   CRLF[0] := CR;
   CRLF[1] := LF;
   defaultBufferSize := BUFFERSIZE;

END FileIO.
