IMPLEMENTATION MODULE RandomGen;

(* Zufallszahlengenerator
   (Knuth, Art of Computer Programming Vol.2, 2nd Ed., pp.16/102)

   jm 15.6.87
   
   tt 25.9.88: RandomCard (x,MaxCard) ergibt keinen berlauf mehr.
   tt 14.3.90: RandomCard berechnet nun direkt aus Rnd() mit Int-Arithm.
               statt ber Random() mit Reals.
   tt 20.5.91: Rnd() zerstrt nicht mehr D3
*)
(*
 * 28.09.1996  -  angepat fr die DOSe (reines M2)
 * 30.09.1996  -  Bugfix
 *)

IMPORT   ModReg;


CONST  a = 1664525;     (* Knuth S.102 Zeile 26 *)
       c = 117L;        (* teilerfremd mit 2^32 fr max. Periode 2^32 *)
  maxRnd = 4294967296.; (* 2^32 fr Normierung auf [0..1) *)

VAR    x: LONGCARD;     (* letzte erzeugte Zufallszahl *)

PROCEDURE Randomize(seed: LONGCARD);
BEGIN
   x := seed;
END Randomize;

PROCEDURE Rnd(): LONGCARD;
VAR   result:  LONGCARD;
      xl, xh,
      al, ah:  LONGCARD;
BEGIN
   (*
    * berechne a * x: drei Teilprodukte bilden
    * (x.high * a.high brauchen wir nicht, fllt dem MOD zum Opfer)
    *)
   ah := a DIV 65536;
   al := a MOD 65536;
   xh := x DIV 65536;
   xl := x MOD 65536;
   x := ((al*xl)+(((al*xh) MOD 65536)*65536)+(((ah*xl) MOD 65536)*65536)+c);
   RETURN x;
   (* ob das nicht nen berlauf gibt? *)
END Rnd;

PROCEDURE RandomCard(min, max: CARDINAL): CARDINAL;
(* liefert Pseudo-Zufallszahl  min <= RandomCard () <= max *)
VAR   result:  CARDINAL;
      rnd:     LONGCARD;
BEGIN
   rnd := Rnd();
   result := SHORT(rnd MOD LONG(max-min+1));
   INC(result, min);
   RETURN result;
END RandomCard;

BEGIN
   ModReg.Register('RandomGen', 0,0,' ',1996,10,16,PROC(NIL));


END RandomGen.